# -*- coding:utf-8 -*-

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Created Date: Monday, July 8th 2019, 8:27:07 am
# Copyright: Tommi Hyppänen


bl_info = {
    "name": "Leafig",
    "category": "Paint",
    "description": "Create textured cutouts out of textures",
    "author": "Tommi Hyppänen (ambi)",
    "location": "Image Editor > Side Panel > Image",
    # "documentation": "",
    "version": (1, 0, 11),
    "blender": (2, 80, 0),
}


# Blender included libraries
import numpy as np
import bpy
import bmesh
import mathutils as mu
from mathutils.geometry import tessellate_polygon

import time
import importlib

if "geo" in locals():
    importlib.reload(geo)
else:
    from . import geo

# if "ilib" in locals():
#     importlib.reload(ilib)
# else:
#     from . import image as ilib

if "ilib" in locals():
    importlib.reload(ilib)
else:
    from .bpy_amb import image as ilib

if "marching" in locals():
    importlib.reload(marching)
else:
    from . import marching

if "chainlib" in locals():
    importlib.reload(chainlib)
else:
    from . import chain as chainlib


# disable assert: __debug__ = False, python -O script.py


class PG_Leafig(bpy.types.PropertyGroup):
    channel: bpy.props.EnumProperty(
        items=[
            ("A", "Alpha", "Extract polygons from alpha channel", 3),
            # ("R", "Red", "", "", 0),
            # ("G", "Green", "", "", 1),
            # ("B", "Blue", "", "", 2),
            # ("L", "Luminance", "", "", 4),
            ("H", "Hue", "Extract polygons using background color hue", 5),
            # ("S", "Saturation", "", "", 6),
            ("D", "Distance", "Extract polygons using distance to background color", 7),
        ],
        name="Channel",
        default="A",
        description="Image channel to read the polygons from",
    )

    back_color: bpy.props.FloatVectorProperty(
        name="Background",
        subtype="COLOR",
        default=(0.0, 0.0, 0.0),
        min=0.0,
        max=1.0,
        description="Background color that is used for visibility calculation.\n"
        "Affects everything else except alpha.\n"
        "Default is black",
    )

    origin: bpy.props.EnumProperty(
        items=[
            ("LEFT", "Left", "", 0),
            ("RIGHT", "Right", "", 1),
            ("BOTTOM", "Bottom", "", 2),
            ("TOP", "Top", "", 3),
            ("CENTER", "Center", "", 4),
        ],
        name="Origin position",
        default="BOTTOM",
        description="Location of the created object origin(s)",
    )

    smooth: bpy.props.IntProperty(
        name="Smoothing", min=0, max=100, default=0, description="Smooth source pixels"
    )

    cutoff: bpy.props.FloatProperty(
        name="Cutoff",
        min=0.0001,
        max=0.9999,
        default=0.5,
        description="Threshold in which the selected channel is considered visible",
    )

    margin: bpy.props.FloatProperty(
        name="Pixel error",
        min=0.0,
        max=100.0,
        default=1.0,
        description="The amount of polygon simplification.\n"
        "How many pixels the polygon edge can move from the border.\n\n"
        "Lower number = more polygons, better accuracy.\n"
        "Higher number = less polygons, lower accuracy.\n\n"
        "Too high of a value can cause self intersections and polygon errors\n"
        "Possible fixes for these errors include setting pixel error to zero\n"
        "or increasing the smoothing value",
    )

    inset: bpy.props.FloatProperty(
        name="Inset", min=0.0, max=50.0, default=0.0, description="Inset the created object border"
    )

    solid_inset: bpy.props.BoolProperty(
        name="Solid inset",
        default=False,
        description="Inset area is created as second material instead of removed",
    )

    separate: bpy.props.BoolProperty(
        name="Separate individuals",
        default=True,
        description="Separate individual parts as their own objects",
    )

    ignore_inner: bpy.props.BoolProperty(
        name="Ignore inner",
        default=False,
        description="Ignore all polygon holes, only create from the outermost loops",
    )

    only_edges: bpy.props.BoolProperty(
        name="Only edges",
        default=False,
        description="Create only polygon edges without filling faces",
    )


def draw_interface(self, context, layout, prg):
    # print(prg.__annotations__.keys())

    def spacer(inpl):
        row = inpl.row()
        row.ui_units_y = 0.5
        row.label(text="")
        return row

    # pixel
    col = layout.box()
    col = col.column(align=True)

    row = col.row()
    row.label(text="Pixels", icon="NODE_TEXTURE")

    spacer(col)

    row = col.row(align=True)
    row.label(text="Channel")
    row.prop(prg, "channel", text="")

    if prg.channel != "A":
        row = col.row(align=True)
        row.label(text="Background")
        row.prop(prg, "back_color", text="")

    spacer(col)

    row = col.row(align=True)
    row.prop(prg, "smooth")

    row = col.row(align=True)
    row.prop(prg, "cutoff")

    # poly
    col = layout.box()
    col = col.column(align=True)
    row = col.row()
    row.label(text="Polygons", icon="SEQ_CHROMA_SCOPE")

    spacer(col)

    row = col.row(align=True)
    row.label(text="Origin")
    row.prop(prg, "origin", text="")

    spacer(col)

    row = col.row(align=True)
    row.prop(prg, "margin")

    row = col.row(align=True)
    row.prop(prg, "inset")

    spacer(col)

    if not prg.only_edges:
        col.row().prop(prg, "separate")
        col.row().prop(prg, "solid_inset")
        col.row().prop(prg, "ignore_inner")
    col.row().prop(prg, "only_edges")


def poly_from_groups(bm, g, multiplier, ensure=False):
    inp = [[mu.Vector(v) for v in c.verts] for c in g]
    polys = tessellate_polygon(inp)

    # write to bmesh
    sverts = []
    for l in inp:
        for v in l:
            loc = (v[1] * multiplier, 0.0, v[0] * multiplier)
            sverts.append(bm.verts.new(loc))

    bm.verts.ensure_lookup_table()
    if ensure:
        ff = set([])
        for p in polys:
            # ensure no dupe faces
            ids = frozenset(tuple(sverts[i].index for i in p))
            if ids not in ff:
                ff.add(ids)
                bm.faces.new([sverts[i] for i in p])
            # else:
            #     print("Leafig: Dupe face.")
    else:
        for p in polys:
            bm.faces.new([sverts[i] for i in p])


def edges_from_groups(bm, g, multiplier):
    new_verts = []
    for chain in g:
        if not chain.invalid:
            cbm = []
            for v in chain.verts:
                nv = bm.verts.new((v[1] * multiplier, 0.0, v[0] * multiplier))
                cbm.append(nv)
            new_verts.append(cbm)

    bm.verts.ensure_lookup_table()

    for cbm in new_verts:
        # bm.faces.new(cbm)
        if len(cbm) > 2:
            for e in list(zip(cbm, cbm[1:] + [cbm[0]])):
                bm.edges.new(e)
        else:
            bm.edges.new((cbm[0], cbm[1]))


def location_remap_XZ(loc, w, h, wdir, hdir):
    zv = mu.Vector([0.0, 0.0, 0.0])
    zv[w] = loc[0] * wdir
    zv[h] = loc[2] * hdir
    return zv


def axis_remap_XZ(bm, w, h, wdir, hdir):
    for v in bm.verts:
        v.co = location_remap_XZ(v.co, w, h, wdir, hdir)

def format_image(image, channel, back_color, verbose=False):
    pixels = np.float32(np.array(image.pixels[:]).reshape(image.size[1], image.size[0], 4))

    xdim = pixels.shape[0]
    ydim = pixels.shape[1]
    mindim = min(xdim, ydim)

    # convert from linear to srgb
    bcol = np.array([ilib.linear2srgb(i) for i in back_color], dtype=np.float32)

    d_RGBA = {"R": 0, "G": 1, "B": 2, "A": 3}
    if channel in d_RGBA:
        pixels[:, :, :3] = np.abs(pixels[:, :, :3] - bcol)
        n = d_RGBA[channel]
        res = pixels[:, :, n]

    d_HSL = {"H": 0, "S": 1, "L": 2}
    if channel in d_HSL:
        chsv = np.array(ilib.rgb2hsv(bcol.reshape((1, 1, 3))), dtype=np.float32).reshape(1, 1, 3)
        pixels_hsv = np.array(ilib.rgb2hsv(pixels[:, :, :3]), dtype=np.float32)
        pixels[:, :, :3] = np.abs(pixels_hsv - chsv)
        res = pixels[:, :, d_HSL[channel]]

    if channel == "D":
        pixels = pixels[:, :, :3].astype(np.float32) - bcol
        res = np.sum(np.abs(pixels), axis=2)

    return res, mindim, xdim, ydim


def run_cutout(
    self,
    context,
    image=None,
    prg=None,
    verbose=False,
    skip_material=False,
    move_to_center=False,
    object_name="Cutout",
    mesh_name="leafig_mesh",
    width_axis="X",
    height_axis="Z",
    width_axis_reverse=False,
    height_axis_reverse=False,
):
    if image is None:
        print("Leafig: Could not find input image")
        self.report({"INFO"}, "Leafig: Could not find input image")
        return {"CANCELLED"}

    if prg is None:
        print("Leafig: Variable group is undefined")
        self.report({"INFO"}, "Leafig: Variable group is undefined")
        return {"CANCELLED"}

    if width_axis == height_axis:
        print("Leafig: Axii must be not the same")
        self.report({"INFO"}, "Leafig: Axii must be not the same")
        return {"CANCELLED"}

    assert width_axis in "XYZ"
    assert height_axis in "XYZ"

    W_m = {"X": 0, "Y": 1, "Z": 2}[width_axis]
    H_m = {"X": 0, "Y": 1, "Z": 2}[height_axis]

    W_dir = -1.0 if width_axis_reverse else 1.0
    H_dir = -1.0 if height_axis_reverse else 1.0

    wm = context.window_manager

    tot = 1000
    wm.progress_begin(0, tot)
    wm.progress_update(100)
    t0 = time.time()

    res, mindim, xdim, ydim = format_image(image, prg.channel, prg.back_color)

    wm.progress_update(200)
    if verbose:
        print("image conversion: {:.2f} sec".format(time.time() - t0))
    t0 = time.time()

    wm.progress_update(300)

    if prg.smooth > 0:
        sval = 1 + prg.smooth * 2
        krn = np.ones(sval) / sval
        f_krn = lambda m: np.convolve(m, krn, mode="same")
        res = np.apply_along_axis(f_krn, axis=1, arr=res)
        res = np.apply_along_axis(f_krn, axis=0, arr=res)
        if verbose:
            print("smoothing: {:.2f} sec".format(time.time() - t0))
        t0 = time.time()

    wm.progress_update(400)

    l_verts = marching.lines_marching(res, prg.cutoff)

    if verbose:
        print("pixel logic: {:.2f} sec".format(time.time() - t0))
    t0 = time.time()

    wm.progress_update(500)

    try:
        chains = chainlib.parse_segments(l_verts)
    except chainlib.ChainException as e:
        print(e)
        self.report({"INFO"}, str(e))
        return {"CANCELLED"}

    if verbose:
        print("parse: {:.2f} sec".format(time.time() - t0))
    t0 = time.time()

    wm.progress_update(600)

    if not chains or len(chains) == 0:
        print("Leafig: No polygons found at this cutoff/smoothing values")
        self.report({"INFO"}, "Leafig: No polygons found at this cutoff/smoothing values")
        return {"CANCELLED"}

    # biggest chain first
    chains = sorted(chains, key=lambda x: -len(x))

    # find inner/outer polygons, use <nm> bools
    # for c in chains:
    #     try:
    #         loc = int(c.verts[c._miny_id][0]), int(c.verts[c._miny_id][1])
    #         c.inner = nm[loc[0], loc[1]]
    #     except Exception as e:
    #         print(e)
    #         print(c._miny_id, len(c.verts))
    #         assert False

    # move verts to right location (to match image/pixel/edge)
    for c in chains:
        for i, v in enumerate(c.verts):
            c.verts[i] = (c.verts[i][0] + 1.0, c.verts[i][1] + 1.0)

    # build polys
    for chain in chains:
        # if not prg.sharp_pixels:
        # chain.smooth(0.5, 2)
        if prg.margin > 0.0:
            chain.simplify(prg.margin)

    if verbose:
        print("num chains:", len(chains))
        print("simplify: {:.2f} sec".format(time.time() - t0))

    t0 = time.time()

    wm.progress_update(700)

    if prg.only_edges == False:
        # go through chain, is it inside any other chain?
        # if not, it's the top chain in the hierarchy
        # make a list of these types of chains and chains inside them

        # TODO: grid search, dict[tuple] = list(chain ids)

        chains = [c for c in chains if not c.invalid]

        if len(chains) == 0:
            print("Leafig: No polygons found at this cutoff/pixel error values")
            self.report({"INFO"}, "Leafig: No polygons found at this cutoff/pixel error values")
            return {"CANCELLED"}

        # find outer loops
        roots = []
        lch = len(chains)
        for ci, c in enumerate(chains):
            wm.progress_update(800 + 100 * ci // lch)
            inside = False
            for i in chains:
                if c == i:
                    continue
                if c.inside(i):
                    inside = True
                    break

            # if chain is not inside any other, mark it as root
            if not inside:
                roots.append(c)

        if prg.ignore_inner == False:
            # find inner loops
            inner = set(chains) - set(roots)
            if verbose:
                print("inner:", len(inner))

            inner_ids = [list() for _ in range(len(roots))]
            # TODO: doesn't check which is inside which
            for i, r in enumerate(roots):
                wm.progress_update(900 + 100 * i // len(roots))
                removal = []
                for c in inner:
                    if r == c:
                        continue
                    if c.inside(r):
                        inner_ids[i].append(c)
                        removal.append(c)

                for r in removal:
                    inner.discard(r)

            groups = [[c] + inner_ids[i] for i, c in enumerate(roots)]
        else:
            groups = [[c] for i, c in enumerate(roots)]

        if verbose:
            print("groups: {:.2f} sec".format(time.time() - t0))
        t0 = time.time()

    def mesh_from_groups(n_func, root, scale):
        # create mesh
        mesh = bpy.data.meshes.new(mesh_name)
        obj = bpy.data.objects.new(object_name, mesh)
        context.collection.objects.link(obj)
        mesh = obj.data
        bm = bmesh.new()

        n_func(bm, root, scale)

        bmesh.ops.beautify_fill(bm, faces=bm.faces[:], edges=bm.edges[:])

        # spine, e_len = geo.build_spine(bm, decimation=1)

        # inset mesh
        if prg.inset > 0.0:
            for f in bm.faces:
                f.select = True

            bmesh.ops.inset_region(
                bm,
                faces=bm.faces[:],
                use_boundary=True,
                use_even_offset=False,
                use_relative_offset=False,
                use_edge_rail=False,
                depth=0.0,
                use_outset=False,
                use_interpolate=True,
                thickness=prg.inset / mindim,
            )

            if not prg.solid_inset:
                bmesh.ops.delete(bm, geom=[f for f in bm.faces if not f.select], context="FACES")
            else:
                for f in bm.faces:
                    f.material_index = 0 if f.select else 1

        # geo.puffer_fish(bm, spine, e_len)

        # select spine

        # for e in border:
        #     e.select = True

        # write UV
        uv_layer = bm.loops.layers.uv.verify()
        for f in bm.faces:
            for vert, loop in zip(f.verts, f.loops):
                loop[uv_layer].uv = (
                    (vert.co[0] * mindim - 0.5) / ydim,
                    (vert.co[2] * mindim - 0.5) / xdim,
                )

        # move object origin
        if prg.origin != "IMAGE_CENTER":
            ploc = root[0].get_end_location(prg.origin)
            offset = mu.Vector([-ploc[1], 0.0, -ploc[0]]) / mindim
        else:
            offset = mu.Vector([-ydim / 2 - 0.5, 0.0, -xdim / 2 - 0.5]) / mindim
        for v in bm.verts:
            v.co += offset

        axis_remap_XZ(bm, W_m, H_m, W_dir, H_dir)

        # make the bmesh the object's mesh
        bm.to_mesh(mesh)
        bm.free()

        if not move_to_center:
            offset = location_remap_XZ(offset, W_m, H_m, W_dir, H_dir)
            obj.location = -offset

        # selecting the object might disrupt trying to determine correct border values
        # obj.select_set(True)
        # context.view_layer.objects.active = obj

        if not skip_material:
            # if prg.create_material:
            mat = ilib.add_material("lfg_" + image.name, image)
            obj.data.materials.append(mat)
            if prg.solid_inset:
                mat2 = ilib.add_material("lfgi_" + image.name, image)
                obj.data.materials.append(mat2)

    if not prg.only_edges:
        if prg.separate:
            # Separate individual root nodes
            for g in groups:
                mesh_from_groups(poly_from_groups, g, 1.0 / mindim)
        else:

            def loop_wrap(bm, x, t):
                for g in groups:
                    poly_from_groups(bm, g, 1.0 / mindim)

            mesh_from_groups(loop_wrap, groups[0], 1.0 / mindim)
    else:
        mesh_from_groups(edges_from_groups, chains, 1.0 / mindim)

    if verbose:
        print("tessellate: {:.2f} sec".format(time.time() - t0))
    t0 = time.time()

    # bmesh.ops.triangulate(
    #     bm, faces=bm.faces, quad_method="BEAUTY", ngon_method="BEAUTY"
    # )

    # print("Output")
    # for x in range(image.shape[0]):
    #     for y in range(image.shape[1]):
    #         image[x, y, 0] += nm[x // st, y // st] * 0.2
    #         image[x, y, 3] = nm[x // st, y // st] * 1.0

    # return image
    wm.progress_end()


class IMAGE_OT_Leafig_Cutout_Script(bpy.types.Operator):
    bl_options = {"REGISTER", "UNDO"}
    bl_idname = "image.leafig_script"
    bl_label = "Leafig Cutout for Scripting"
    bl_description = "Creates a textured cutout from image"

    image_name: bpy.props.StringProperty(name="Image name", default="")

    object_name: bpy.props.StringProperty(name="Object name", default="")

    print_info: bpy.props.BoolProperty(
        name="Print info", default=False, description="Prints some debug info"
    )

    channel: bpy.props.EnumProperty(
        items=[
            ("A", "Alpha", "Extract polygons from alpha channel", 3),
            ("H", "Hue", "Extract polygons using background color hue", 5),
            ("D", "Distance", "Extract polygons using distance to background color", 7),
        ],
        name="Channel",
        default="A",
        description="Image channel to read the polygons from",
    )

    width_axis: bpy.props.EnumProperty(
        items=[
            ("X", "X-Axis", "X-Axis", 0),
            ("Y", "Y-Axis", "Y-Axis", 1),
            ("Z", "Z-Axis", "Z-Axis", 2),
        ],
        name="Width axis",
        default="X",
        description="The width axis for the 2D to 3D mapping",
    )

    width_axis_reverse: bpy.props.BoolProperty(
        name="Width axis revers", default=False, description="Reverse width axis direction"
    )

    height_axis: bpy.props.EnumProperty(
        items=[
            ("X", "X-Axis", "X-Axis", 0),
            ("Y", "Y-Axis", "Y-Axis", 1),
            ("Z", "Z-Axis", "Z-Axis", 2),
        ],
        name="Height axis",
        default="Z",
        description="The height axis for the 2D to 3D mapping",
    )

    height_axis_reverse: bpy.props.BoolProperty(
        name="Height axis revers", default=False, description="Reverse height axis direction"
    )

    back_color: bpy.props.FloatVectorProperty(
        name="Background",
        subtype="COLOR",
        default=(0.0, 0.0, 0.0),
        min=0.0,
        max=1.0,
        description="Background color that is used for visibility calculation.\n"
        "Affects everything else except alpha.\n"
        "Default is black",
    )

    origin: bpy.props.EnumProperty(
        items=[
            ("LEFT", "Left", "", 0),
            ("RIGHT", "Right", "", 1),
            ("BOTTOM", "Bottom", "", 2),
            ("TOP", "Top", "", 3),
            ("CENTER", "Center", "", 4),
            ("IMAGE_CENTER", "Imagecenter", "", 5),
        ],
        name="Origin position",
        default="BOTTOM",
        description="Location of the created object origin(s)",
    )

    smooth: bpy.props.IntProperty(
        name="Smoothing", min=0, max=100, default=0, description="Smooth source pixels"
    )

    cutoff: bpy.props.FloatProperty(
        name="Cutoff",
        min=0.0001,
        max=0.9999,
        default=0.5,
        description="Threshold in which the selected channel is considered visible",
    )

    margin: bpy.props.FloatProperty(
        name="Pixel error",
        min=0.0,
        max=100.0,
        default=1.0,
        description="The amount of polygon simplification.\n"
        "How many pixels the polygon edge can move from the border.\n\n"
        "Lower number = more polygons, better accuracy.\n"
        "Higher number = less polygons, lower accuracy.\n\n"
        "Too high of a value can cause self intersections and polygon errors\n"
        "Possible fixes for these errors include setting pixel error to zero\n"
        "or increasing the smoothing value",
    )

    inset: bpy.props.FloatProperty(
        name="Inset", min=0.0, max=50.0, default=0.0, description="Inset the created object border"
    )

    solid_inset: bpy.props.BoolProperty(
        name="Solid inset",
        default=False,
        description="Inset area is created as second material instead of removed",
    )

    separate: bpy.props.BoolProperty(
        name="Separate individuals",
        default=True,
        description="Separate individual parts as their own objects",
    )

    ignore_inner: bpy.props.BoolProperty(
        name="Ignore inner",
        default=False,
        description="Ignore all polygon holes, only create from the outermost loops",
    )

    only_edges: bpy.props.BoolProperty(
        name="Only edges",
        default=False,
        description="Create only polygon edges without filling faces",
    )

    skip_material: bpy.props.BoolProperty(
        name="Skip material", default=False, description="Don't write UV or material"
    )

    move_to_center: bpy.props.BoolProperty(
        name="Move to center", default=False, description="Move resulting cutouts to origin"
    )

    def execute(self, context):
        if self.image_name not in bpy.data.images:
            self.report({"INFO"}, "Leafig: Image not found")
            return {"CANCELLED"}
        else:
            image = bpy.data.images[self.image_name]

        if self.width_axis == self.height_axis:
            self.report({"INFO"}, "Leafig: Width and height axis must be different")
            return {"CANCELLED"}

        run_cutout(
            self,
            context,
            image,
            self,
            self.print_info,
            self.skip_material,
            self.move_to_center,
            self.object_name if len(self.object_name) > 0 else "Cutout",
            width_axis=self.width_axis,
            height_axis=self.height_axis,
            width_axis_reverse=self.width_axis_reverse,
            height_axis_reverse=self.height_axis_reverse,
        )
        return {"FINISHED"}


class IMAGE_OT_Leafig_Cutout(bpy.types.Operator):
    bl_options = {"REGISTER", "UNDO"}
    bl_idname = "image.leafig_cutout"
    bl_label = "Leafig Cutout"
    bl_description = "Creates a textured cutout from image"

    # TODO: calculate proper origin position for object when no separate individuals
    # TODO: also test the second step in triangle strip self collision
    # TODO: marching squares also parse
    # TODO: replace inner detection with integer count based on 2d normals

    prg: bpy.props.PointerProperty(type=PG_Leafig)

    def invoke(self, context, event):
        # copy from panel to operator
        for k in self.prg.__annotations__.keys():
            setattr(self.prg, k, getattr(context.scene.leafig, k))
        return self.execute(context)

    def draw(self, context):
        draw_interface(self, context, self.layout, self.prg)

    def execute(self, context):
        try:
            image = ilib.get_teximage(context)
        except AttributeError:
            self.report({"INFO"}, "Leafig: The operator needs to be run from inside image editor")
            return {"CANCELLED"}

        run_cutout(self, context, image, self.prg)
        return {"FINISHED"}


class IMAGE_OT_Leafig_Rebuild_UV(bpy.types.Operator):
    bl_options = {"REGISTER", "UNDO"}
    bl_idname = "image.leafig_rebuild_uv"
    bl_label = "Leafig Rebuild UV"
    bl_description = "Rebuild  UV based on mesh vertex locations"

    def execute(self, context):
        obj = context.active_object

        if obj == None:
            print("Leafig: No active object found")
            self.report({"INFO"}, "Leafig: No active object found")
            return {"CANCELLED"}

        me = obj.data

        try:
            image = ilib.get_teximage(context)
        except AttributeError:
            self.report({"INFO"}, "Leafig: The operator needs to be run from inside image editor")
            return {"CANCELLED"}

        if image == None:
            print("Leafig: No image selected in the editor")
            self.report({"INFO"}, "Leafig: No image selected in the editor")
            return {"CANCELLED"}

        bm = bmesh.new()
        bm.from_mesh(me)

        xdim = image.size[1]
        ydim = image.size[0]
        mindim = min(xdim, ydim)

        loc = obj.matrix_world.translation
        print(loc)

        # TODO: base uv on material image
        uv_layer = bm.loops.layers.uv.verify()
        for f in bm.faces:
            for vert, loop in zip(f.verts, f.loops):
                loop[uv_layer].uv = (
                    ((vert.co[0] + loc.x) * mindim - 0.5) / ydim,
                    ((vert.co[2] + loc.z) * mindim - 0.5) / xdim,
                )

        bm.to_mesh(me)
        bm.free()

        return {"FINISHED"}


class IMAGE_OT_Leafig_Surfacer(bpy.types.Operator):
    bl_options = {"REGISTER", "UNDO"}
    bl_idname = "image.leafig_surfacer"
    bl_label = "Leafig Build Surface"
    bl_description = "Build a polygonal surface based on connected edge loop"

    def execute(self, context):
        obj = context.active_object

        if obj == None:
            print("Leafig: No active object found")
            self.report({"INFO"}, "Leafig: No active object found")
            return {"CANCELLED"}

        me = obj.data
        bm = bmesh.new()
        bm.from_mesh(me)

        chains = chainlib.from_mesh(bm)
        print("Surfacer: Chain count:", len(chains))

        # for ch in chains:
        ch = chains[0]
        normals = np.array(ch.normals())
        for i, n in enumerate(normals):
            vloc = ch.verts[i]
            nloc = vloc + n * 0.005
            bm.verts.new((nloc[0], 0.0, nloc[1]))

        bm.to_mesh(me)
        bm.free()

        print("Surfacer finished.")

        return {"FINISHED"}


class IMAGE_PT_Leafig(bpy.types.Panel):
    bl_label = "Leafig"
    bl_space_type = "IMAGE_EDITOR"
    bl_region_type = "UI"
    bl_category = "Image"

    def draw(self, context):
        # layout = self.layout.column(align=True)
        layout = self.layout
        row = layout.row()
        row.scale_y = 1.5
        row.operator(IMAGE_OT_Leafig_Cutout.bl_idname, text="Create Cutout")
        draw_interface(self, context, layout, context.scene.leafig)

        row = layout.row()
        row.label(text="Corrective ops:")
        row = layout.row()
        row.operator(IMAGE_OT_Leafig_Rebuild_UV.bl_idname, text="Rebuild UV")
        # row = layout.row()
        # row.operator(IMAGE_OT_Leafig_Surfacer.bl_idname, text="Surfacer")


classes = (
    PG_Leafig,
    IMAGE_OT_Leafig_Cutout,
    IMAGE_OT_Leafig_Cutout_Script,
    IMAGE_OT_Leafig_Rebuild_UV,
    IMAGE_OT_Leafig_Surfacer,
    IMAGE_PT_Leafig,
)


def register():
    for c in classes:
        bpy.utils.register_class(c)
    bpy.types.Scene.leafig = bpy.props.PointerProperty(type=PG_Leafig)


def unregister():
    for c in classes[::-1]:
        bpy.utils.unregister_class(c)
    del bpy.types.Scene.leafig
