import numpy as np
import math
import mathutils as mu
from collections import defaultdict


class ChainException(Exception):
    """ Exception for Chain class """

    pass


def _dpcall3(chain, eps):
    stack = [(0, len(chain) - 1)]
    result = []
    while stack:
        pl0, pl1 = stack.pop()

        dmax = 0
        index = 0

        x1, y1 = chain[pl0]
        x2, y2 = chain[pl1]
        y21 = y2 - y1
        x21 = x2 - x1
        tsq2 = math.sqrt(y21 ** 2 + x21 ** 2)
        xyyx = x2 * y1 - y2 * x1
        for i in range(pl0 + 1, pl1):
            x0, y0 = chain[i]
            d = 0.0

            # if x1 == x2 and y1 == y2:
            #     d = math.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)
            # else:
            d = abs(y21 * x0 - x21 * y0 + xyyx) / tsq2

            if d > dmax:
                index = i
                dmax = d

        if dmax <= eps:
            result.append(chain[pl0])
        else:
            stack.append((index, pl1))
            stack.append((pl0, index - 1))

    # result.append(chain[-1])
    return result


def douglas_peucker(ch, margin):
    # t = ch[:]
    # res = _dpcall2(ch, margin)
    # assert len(res) <= len(t)
    return _dpcall3(ch, margin)


class Chain:
    def __init__(self):
        self.verts = []
        self.inner = False
        self.bm_verts = []
        self.invalid = False

        self._pvert = None
        self._cvert = None

        self._minx = None
        self._minx_id = 0

        self._miny = None
        self._miny_id = 0

        self._maxx = None
        self._maxx_id = 0

        self._maxy = None
        self._maxy_id = 0

        self._segments = None
        self._np_segments = None

        self.parent = None
        self.children = []

    def __len__(self):
        return len(self.verts)

    def verify(self):
        for i in range(len(self.verts)):
            v0 = self.verts[i]
            v1 = self.verts[(i + 1) % len(self.verts)]
            if v0 == v1:
                return False
        return True

    def add_vert(self, vert):
        if self._cvert == None:
            self._pvert = vert
            self._cvert = vert
            self._miny = vert[1]
            self._minx = vert[0]
            self._maxy = vert[1]
            self._maxx = vert[0]

        self._pvert = self._cvert
        self._cvert = vert

        if self._cvert[0] < self._minx:
            self._minx = self._cvert[0]
            self._minx_id = len(self.verts)

        if self._cvert[1] < self._miny:
            self._miny = self._cvert[1]
            self._miny_id = len(self.verts)

        if self._cvert[0] > self._maxx:
            self._maxx = self._cvert[0]
            self._maxx_id = len(self.verts)

        if self._cvert[1] > self._maxy:
            self._maxy = self._cvert[1]
            self._maxy_id = len(self.verts)

        self.verts.append(vert)

    def get_end_location(self, loc="BOTTOM"):
        if loc == "CENTER":
            # x, y = 0.0, 0.0
            # for v in self.verts:
            #     x += v[0]
            #     y += v[1]
            # return (x / len(self.verts), y / len(self.verts))
            maxx = self.verts[self._maxx_id][0]
            minx = self.verts[self._minx_id][0]
            maxy = self.verts[self._maxy_id][1]
            miny = self.verts[self._miny_id][1]
            return ((maxx + minx) / 2, (maxy + miny) / 2)

        return self.verts[
            {
                "LEFT": self._miny_id,
                "RIGHT": self._maxy_id,
                "BOTTOM": self._minx_id,
                "TOP": self._maxx_id,
            }[loc]
        ]

    def _recalc_bb(self):
        self._miny = self.verts[0][1]
        self._minx = self.verts[0][0]
        self._maxy = self.verts[0][1]
        self._maxx = self.verts[0][0]
        self._miny_id = 0
        self._minx_id = 0
        self._maxy_id = 0
        self._maxx_id = 0
        for i in range(len(self.verts)):
            cvert = self.verts[i]
            if cvert[0] < self._minx:
                self._minx = cvert[0]
                self._minx_id = i

            if cvert[1] < self._miny:
                self._miny = cvert[1]
                self._miny_id = i

            if cvert[0] > self._maxx:
                self._maxx = cvert[0]
                self._maxx_id = i

            if cvert[1] > self._maxy:
                self._maxy = cvert[1]
                self._maxy_id = i

    def get_segments(self):
        if self._segments == None:
            self._segments = list(zip(self.verts, list(self.verts[1:]) + [self.verts[0]]))
            self._np_segments = np.array(self._segments)
        return (self._segments, self._np_segments)

    def inside(self, other):
        # if outside the bounding box, obviously false
        if (
            self._minx > other._maxx
            or self._miny > other._maxy
            or self._maxx < other._minx
            or self._maxy < other._miny
        ):
            return False

        # if number of intersects to this test line is odd
        # the point is inside the tested chain
        mp = self.verts[self._miny_id]
        tlp = mp[0] + 0.001
        test_line = ((tlp, other._miny - 1.0), (tlp, mp[1]))
        isect_count = 0

        # ...if segment points at opposite sides of the tested line
        # get the numpy array of segments from other
        np_segs = other.get_segments()[1]
        # minus times minus is plus, etc...
        # only under zero if the sign of sides is not equal
        seg_ids = (np_segs[:, 0, 0] - tlp) * (np_segs[:, 1, 0] - tlp) <= 0.0
        # ...intersect with the test line
        for s in np_segs[seg_ids]:
            if mu.geometry.intersect_line_line_2d(test_line[0], test_line[1], s[0], s[1]):
                isect_count += 1
        return isect_count % 2 != 0

    def simplify(self, error):
        tmp = douglas_peucker(self.verts, error)
        if len(tmp) > 3:
            self.verts = tmp
            self._recalc_bb()
        else:
            self.invalid = True

    def smooth(self, amount, t):
        ch = self.verts
        for _ in range(t):
            for i in range(len(ch)):
                vl = ch[(i - 1) % len(ch)]
                vr = ch[(i + 1) % len(ch)]
                x = ((vl[0] + vr[0]) * 0.5) * amount + ch[i][0] * (1.0 - amount)
                y = ((vl[1] + vr[1]) * 0.5) * amount + ch[i][1] * (1.0 - amount)
                ch[i] = (x, y)
        self.verts = ch

    def normals(self):
        ch = np.array(self.verts)
        normals = np.zeros((ch.shape[0], 2))
        for i in range(ch.shape[0]):
            vp = ch[(i - 1) % ch.shape[0]]
            vecp = vp - ch[i]
            # rot90
            vecp[0], vecp[1] = vecp[1], -vecp[0]
            vecp /= np.linalg.norm(vecp)
            normals[i] = vecp

        # average over previous vert
        normals += np.roll(normals, -1, axis=0)
        for i in range(normals.shape[0]):
            normals[i] /= np.linalg.norm(normals[i])
        return normals


def parse_segments(l_verts, bounds=True):
    # create verts and connect edges
    verts = []
    vert_ids = {}
    vert_edges = defaultdict(list)
    for l in l_verts:
        v0, v1 = l
        ev0, ev1 = 0, 0

        # limit to bounds
        if bounds and (v0[0] < 0 or v0[1] < 0 or v1[0] < 0 or v1[1] < 0):
            continue

        if v0 in vert_ids:
            ev0 = vert_ids[v0]
        else:
            verts.append(v0)
            ev0 = len(verts) - 1
            vert_ids[v0] = ev0

        if v1 in vert_ids:
            ev1 = vert_ids[v1]
        else:
            verts.append(v1)
            ev1 = len(verts) - 1
            vert_ids[v1] = ev1

        vert_edges[ev0].append(ev1)
        vert_edges[ev1].append(ev0)

    # parse chains
    chains = []
    remaining = set(list(range(len(verts))))
    failures = 0
    while len(remaining) > 0:
        chains.append(Chain())
        head = remaining.pop()
        prev = head
        start = head
        chains[-1].add_vert(verts[head])
        while True:
            # connected vert ids
            veh = vert_edges[head]

            if len(veh) == 1:
                head = veh[0]
                remaining.discard(head)
                chains[-1].add_vert(verts[head])
                failures += 1
                # added.add(head)
                break

            if len(veh) > 2:
                chains.pop()
                remaining.discard(head)
                failures += 1
                break

            here = veh[0] if veh[0] != head and veh[0] != prev else veh[1]
            prev = head
            head = here

            if head == start:
                break

            remaining.discard(head)
            chains[-1].add_vert(verts[head])

        # dist = np.linalg.norm(np.array(verts[prev]) - np.array(verts[start]))
        # if dist <= 0.0:
        #     print("too small:", dist)
        #     assert False
        # if dist >= 2.0:
        #     print("too big:", dist)
        #     assert False

        # if chains and (len(chains[-1]) < 6 or head != start):
        if chains and head != start:
            print("Chain failure")
            failures += 1
            chains.pop()

    # TODO: return codes instead
    if len(chains) == 0:
        raise ChainException("在此截止值处找不到多边形，请调整截止值")
        # print("Leafig: No polygons found at this cutoff value")
        # self.report({"INFO"}, "Leafig: No polygons found at this cutoff value")

    if failures > 0:
        raise ChainException("Chain failures: {}".format(failures))
        # print("Leafig: Chain failures:", failures)
        # self.report({"INFO"}, "{} chain failures!".format(failures))

    return chains


def from_mesh(bm):
    segments = []
    for e in bm.edges:
        segments.append([(v.co.x, v.co.z) for v in e.verts])
    # print(segments[:10])
    return parse_segments(segments, bounds=False)
