import numpy as np
import math
import mathutils as mu
from collections import defaultdict
import bmesh

import importlib

if "marching" in locals():
    importlib.reload(marching)
else:
    from . import marching


class VertDistance:
    def __init__(self, dist=None, maxdist=None):
        self.distance = dist
        self.max_distance = maxdist


# def dist_point_line(p, l1, l2):
#     x0, y0 = p
#     x1, y1 = l1
#     x2, y2 = l2
#     y21 = y2 - y1
#     x21 = x2 - x1
#     if l1[0] == l2[0] and l1[1] == l2[1]:
#         return math.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)
#     return abs(y21 * x0 - x21 * y0 + x2 * y1 - y2 * x1) / math.sqrt(y21 ** 2 + x21 ** 2)


def tri_ori(p1, p2, p3):
    # skip colinearity test
    return (p2[1] - p1[1]) * (p3[0] - p2[0]) - (p2[0] - p1[0]) * (p3[1] - p2[1]) > 0


def radial_edges(iv):
    loop = iv.link_loops[0]
    # Return if any edges are non-manifold
    # if not any(e for e in iv.link_edges if not e.is_manifold):
    #     return None
    eg = []
    while True:
        eg.append(loop.edge)
        loop = loop.link_loop_radial_next.link_loop_next
        le = loop.edge
        if le == eg[0]:  # or (le.verts[0] != iv and le.verts[1] != iv):
            break
    # assert len(eg) == len(iv.link_edges)
    return eg


def triangle_strip_intersect(bm, v0, s_edge, limit_len=0.0):
    # tstrip core point
    if not v0.is_valid:
        return True

    # if not (v0 == s_edge.verts[0] or v0 == s_edge.verts[1]):
    #     return True

    # e_idx = [i for i, e in enumerate(v0.link_edges) if e.select]
    # if len(e_idx) == 0:
    #     return True

    # e_idx = e_idx[0]
    # s_edge = v0.link_edges[e_idx]
    ov = s_edge.other_vert(v0)
    ovc = ov.co + (v0.co - ov.co) * 0.001

    edges = radial_edges(v0)
    # if edges is None or len(edges) < 3:
    #     return True
    for e in edges:
        assert v0 == e.verts[0] or v0 == e.verts[1]

    if limit_len > 0.0:
        loc = edges.index(s_edge)
        e_left = edges[(loc + 1) % len(edges)]
        e_right = edges[(loc - 1) % len(edges)]
        avg_len = 0.0
        for e in edges:
            avg_len += e.calc_length()
        avg_len /= len(edges)

        if (e_left.calc_length() + e_right.calc_length()) * limit_len * 0.5 < avg_len:
            return True

    for i, e in enumerate(edges):
        # link_edges isn't in radial order dT_Tb
        # assert len(set(p1v.link_edges) & set(p2v.link_edges)) > 0
        p1 = e.other_vert(v0).co.xz
        p2 = edges[(i + 1) % len(edges)].other_vert(v0).co.xz
        if tri_ori(ovc.xz, p1, p2) != tri_ori(v0.co.xz, p1, p2):
            return True

    return False


def triangle_strip2(bm, v0, ce):
    """ Tstrip based on v0 and connected, selected edge """
    # ce = next(e for e in v0.link_edges if e.select)
    bmesh.ops.pointmerge(bm, verts=ce.verts, merge_co=ce.other_vert(v0).co)
    return True


def vert_distance_to_border(bm, c_faces, v):
    # calc min distance to border

    # check connected faces border edges for minimun distance to border
    b_edges = set()
    for f in c_faces:
        for i in f.edges:
            if i.is_boundary:
                b_edges.add(i)

    dist = 100000.0
    for be in b_edges:
        v0, v1 = be.verts
        pt = mu.geometry.intersect_point_line(v.co, v0.co, v1.co)[0]
        d = (pt - v.co).length * 2.0
        if d < dist:
            dist = d

    return dist


def split_50(bm, edges, border_verts, mv=0.0, proportional=False, edge_lengths=None):
    new_verts = set()
    # vert_seg = {}
    # dist_to_border = {v: 0.0 for v in bm.verts}

    edges_a = set()
    edges_b = set()

    new_locs = {}

    for e in edges:
        # get edge parameters
        # vector from center to border
        tv = e.verts[0] if e.verts[0] not in border_verts else e.verts[1]
        bvert = e.other_vert(tv).co
        dif = bvert - tv.co
        diflen = dif.length
        difnrm = dif.normalized()

        # split edge
        orig_len = e.calc_length()
        ne, v = bmesh.utils.edge_split(e, e.verts[0], 0.5)

        # TODO: closest distance to border through connected faces

        # propagate length data to table
        if edge_lengths is not None:
            e_len = None
            for eb in v.link_edges:
                ov = eb.other_vert(v)
                if ov in edge_lengths:
                    e_len = edge_lengths[ov]
            if e_len == None:
                e_len = orig_len / 2
                e_t = vert_distance_to_border(bm, e.link_faces, v)
                if e_t < e_len:
                    e_len = e_t
            edge_lengths[v] = e_len

        # put edge closest to border into edges_a
        med = None
        if e.verts[0] in border_verts or e.verts[1] in border_verts:
            med = e
            edges_a.add(e)
            edges_b.add(ne)
        else:
            med = ne
            edges_a.add(ne)
            edges_b.add(e)

        new_verts.add(v)

        if mv > 0.0:
            # slide by distance instead of 50/50
            mv_this = mv

            # calc perpendicular direction from the border
            ev = med.other_vert(v)
            evec = (v.co - ev.co).normalized()

            b_edges = [e for e in ev.link_edges if e.is_boundary]
            b_vecs = [(e0.verts[0].co - e0.verts[1].co) for e0 in b_edges]
            b_vecs = [-v0 if b_edges[i].verts[0] == ev else v0 for i, v0 in enumerate(b_vecs)]
            b_lens = [v0.length for v0 in b_vecs]
            b_norms = [v0.normalized() for v0 in b_vecs]
            b_dots = [v0.dot(evec) for v0 in b_norms]
            b_nearest = max(b_dots)
            b_choice = b_dots.index(b_nearest)

            # travel angle (between edge and border) compensation
            mm = 1.0
            if b_nearest > 0:
                mm = 1.0 / math.acos(b_nearest)
                # if the border edge is short, limit distance traveled
                if mm * mv_this > b_lens[b_choice]:
                    mm = b_lens[b_choice] / mv_this
                if mm < 1.0:
                    mm = 1.0
            mv_this *= mm

            # write move
            if mv_this > diflen:
                mv_this = diflen

            new_locs[v] = bvert - difnrm * mv_this

        # el = e.calc_length()
        # dist = el

        # calc min distance to border
        # c_faces = list(e.link_faces)
        # b_edges = []

        # check connected faces border edges for minimun distance to border
        # for f in c_faces:
        #     b_edges.extend([i for i in f.edges if i in border])

        # for be in b_edges:
        #     v0, v1 = be.verts
        #     pt = mu.geometry.intersect_point_line(v.co, v0.co, v1.co)[0]
        #     d = (pt - v.co).length * 2.0
        #     if d < dist:
        #         dist = d

        # vert_seg[v] = e_vec
        # dist_to_border[v] = dist

    for k, v in new_locs.items():
        k.co = v

    return new_verts, edges_a, edges_b


def connect_verts(bm, new_verts):
    f_verts = set()
    f_connect = defaultdict(list)
    new_border = set()
    created_edges = set()

    for f in bm.faces[:]:
        vloop = [v for v in f.verts if v in new_verts]
        if len(vloop) > 1:
            nedges = list(zip(vloop, vloop[1:] + [vloop[0]]))
            if len(nedges) == 2:
                nedges = [nedges[0]]

            # Check for duplicate edges, move to next face if exists
            dupe_exists = False
            for ne in nedges:
                if not tuple(ne) in created_edges:
                    created_edges.add(tuple(ne))
                    created_edges.add(tuple(ne[::-1]))
                else:
                    dupe_exists = True
                    break
            if dupe_exists:
                continue

            bedges = [bm.edges.new((ne[0], ne[1])) for ne in nedges]
            if len(bedges) > 1:
                # add longest edges to be split later
                be_len = [be.calc_length() for be in bedges]
                max_idx = be_len.index(max(be_len))
                # min_idx = be_len.index(min(be_len))

                # TODO: connect all edges to new created center vert instead

                for i in range(len(bedges)):
                    tbe = bedges[i]

                    # add all except the longest edge to the border
                    if max_idx != i:
                        v0, v1 = tbe.verts
                        f_verts.add(v0)
                        f_connect[v1].append(v0)
                        f_verts.add(v1)
                        f_connect[v0].append(v1)
                        new_border.add(tbe)
            else:
                # only 1 edge
                v0, v1 = bedges[0].verts

                f_verts.add(v0)
                f_connect[v1].append(v0)
                f_verts.add(v1)
                f_connect[v0].append(v1)
                new_border.add(bedges[0])

            # split face according to new_verts
            bmesh.utils.face_split_edgenet(f, bedges)
        elif len(vloop) == 1:
            # only 1 vert
            f_verts.add(vloop[0])

            # split quads
            if len(f.verts) == 4:
                vertl = f.loops[list(f.verts).index(vloop[0])]
                vert_a = vloop[0]
                vert_b = vertl.link_loop_next.link_loop_next.vert
                f, _ = bmesh.utils.face_split(f, vert_a, vert_b)
                for e in f.edges:
                    if (e.verts[0] == vert_a and e.verts[1] == vert_b) or (
                        e.verts[1] == vert_a and e.verts[0] == vert_b
                    ):
                        new_border.add(e)
                        break
    return new_border


def smooth_spine(bm, vert_seg, f_verts, f_connect):
    for _ in range(0):
        for v in f_verts:
            c_verts = f_connect[v]

            # check overlap
            # if len(c_verts) == 2:
            #     p00, p01 = [mu.Vector((i.y, i.z)) for i in vert_seg[v]]
            #     p10, p11 = [mu.Vector((i.co.y, i.co.z)) for i in c_verts]
            #     ipt = mu.geometry.intersect_line_line_2d(p00, p01, p10, p11)
            #     if ipt is None:
            #         continue

            if len(c_verts) > 1:
                # regular smoothing
                tt = mu.Vector((0, 0, 0))
                for cv in c_verts:
                    tt += cv.co
                v.co = tt / len(c_verts) * 0.5 + v.co * 0.5


def linked_edges(e, v, test):
    return [ve for ve in v.link_edges if ve != e and ve in test]


def leaf_connection_point(l, s_edges):
    if not l.is_valid:
        return None
    le = linked_edges(l, l.verts[1], s_edges)
    return l.verts[any(le) * 1] if len(le) > 0 else None


def shrink_leafs(leafs, s_edges, discard=True):
    next_run = set()
    for l in leafs:
        if not l.is_valid:
            s_edges.discard(l)
            continue

        # find which end to iterate
        lte = linked_edges(l, l.verts[1], s_edges)

        # get connected edges amount
        # only iterate when one edge connected, not zero or more than one
        le = linked_edges(l, l.verts[any(lte) * 1], s_edges)
        if le and len(le) == 1:
            next_run.add(le[0])
        if discard:
            s_edges.discard(l)
    return next_run


def spine_leafs(spine):
    return set(
        e
        for e in spine
        if any(linked_edges(e, e.verts[0], spine)) == False
        or any(linked_edges(e, e.verts[1], spine)) == False
    )


def decimate_spine(bm, spine, limit, count):
    # iter from leafs
    # parse backwards the spine to remove small bones

    s_edges = set(spine)
    print("spine:", len(spine), len(s_edges))

    # move away one step from the border
    leafs = shrink_leafs(spine_leafs(spine), s_edges)
    print("leafs:", len(leafs))

    for e in bm.edges:
        e.select = False

    for e in s_edges:
        e.select = True

    for _ in range(count):
        if len(leafs) == 0:
            break
        for l in leafs:
            # v = leaf endpoint
            v = l.verts[0] if len(linked_edges(l, l.verts[0], s_edges)) == 0 else l.verts[1]
            v0 = l.other_vert(v)
            if not triangle_strip_intersect(bm, v, l, limit_len=limit):
                bmesh.ops.weld_verts(bm, targetmap={v: v0})

        s_edges = set(e for e in s_edges if e.is_valid)
        leafs = spine_leafs(s_edges)

    return s_edges


def extrude_spine(bm, spine, edge_len, profile=1.0, location=0.0):
    # extrude spine
    spine_verts = set()
    for e in spine:
        spine_verts.add(e.verts[0])
        spine_verts.add(e.verts[1])

    e_amount = (1 - location ** profile) ** (1 / profile)

    for v in spine_verts:
        extrude = 0.0
        if v in edge_len:
            extrude = edge_len[v]
        # if extrude > 0.1:
        #     extrude = 0.1
        v.co.y = -e_amount * extrude


def calc_border(bm):
    border = set([e for e in bm.edges if e.is_boundary])
    border_verts = set()
    for e in border:
        border_verts.add(e.verts[0])
        border_verts.add(e.verts[1])

    return border, border_verts


class River:
    def __init__(self, bm, spine):
        self.spine = spin
        self.upstream = {}
        self.downstream = defaultdict(list)


def build_spine(bm, decimation=2):
    border, border_verts = calc_border(bm)

    # split non-boundary edges, move verts by edge_len / 2
    e_len = {}
    inner_edges = [e for e in bm.edges if not e.is_boundary]
    new_verts, edges_a, edges_b = split_50(bm, inner_edges, border_verts, edge_lengths=e_len)

    # connect the new verts from the previous edge splits
    raw_spine = connect_verts(bm, new_verts)
    spine = decimate_spine(bm, raw_spine, 0.0, decimation)
    # spine = decimate_spine(bm, spine, 1.2, 5)
    # spine = decimate_spine(bm, raw_spine, 0.0, 0)

    # rotate all edges to touch the border
    for e in set(bm.edges) - spine - border:
        if not (e.verts[0].is_boundary or e.verts[1].is_boundary):
            bmesh.ops.rotate_edges(bm, edges=[e])

    return spine, e_len

    # while tree:
    #     count = 0
    #     for e in list(tree):
    #         for i in range(2):
    #             if e.verts[i] in leafs:
    #                 v0 = e.verts[1 - i]
    #                 if sum(1 for e in v0.link_edges if e.select) == 2:
    #                     leafs.add(e.verts[1 - i])
    #                 leafs.discard(e.verts[i])
    #                 e.select = False
    #                 tree.discard(e)
    #                 faces.extend(e.verts[i].link_faces)
    #                 count += 1
    #     if count == 0:
    #         break

    # for f in faces:
    #     f.select = True

    # traversed = set()
    # for lv in leafs:
    #     traversed.add(lv)
    #     lv.select = True
    # lv = [e.other_vert(lv) for e in lv.link_edges if e.select][0]
    # traversed.add(lv)
    # while True:
    #     c_edges = [e for e in lv.link_edges if e.select == True]
    #     if len(c_edges) != 2:
    #         break
    #     lv = [v for v in [e.other_vert(lv) for e in c_edges] if v not in traversed]
    #     if len(lv) == 0:
    #         break
    #     lv = lv[0]
    #     traversed.add(lv)


def puffer_fish(bm, spine, e_len):
    border, border_verts = calc_border(bm)
    spine_verts = set()
    for e in spine:
        spine_verts.add(e.verts[0])
        spine_verts.add(e.verts[1])

    # check KD-tree for minimal distances to border and overwrite if less
    kd = mu.kdtree.KDTree(len(border_verts))
    for i, v in enumerate(border_verts):
        kd.insert(v.co, i)
    kd.balance()

    for v in spine_verts:
        _, di, vd = kd.find(v.co)
        if vd < e_len[v]:
            e_len[v] = vd

    # limit thickness
    for v in e_len.keys():
        if e_len[v] > 0.09:
            e_len[v] = 0.09

    # extrude_spine(bm, spine, e_len, location=0.0)

    new_edges = set(bm.edges) - spine - border
    # new_verts, edges_a, edges_b = split_50(bm, new_edges, border_verts, mv=0.05, edge_lengths=e_len)
    new_verts, edges_a, edges_b = split_50(bm, new_edges, border_verts, edge_lengths=e_len)
    border0 = connect_verts(bm, new_verts)

    # new_verts, edges_a, _ = split_50(bm, edges_a, border_verts, edge_lengths=e_len)
    # border1 = connect_verts(bm, new_verts)
    # new_verts, _, _ = split_50(bm, edges_b, border_verts, edge_lengths=e_len)
    # border2 = connect_verts(bm, new_verts)
    # new_verts, edges_a, edges_b = split_50(bm, edges_a, border_verts, edge_lengths=e_len)
    # border3 = connect_verts(bm, new_verts)

    # prof = 2.5

    # extrude the intermediates
    # extrude_spine(bm, border0, e_len, location=1 / 2, profile=prof)
    # extrude_spine(bm, border1, e_len, location=3 / 4, profile=prof)
    # extrude_spine(bm, border2, e_len, location=1 / 4, profile=prof)
    # extrude_spine(bm, border3, e_len, location=7 / 8, profile=prof)

    # mark spine
    for f in bm.faces:
        f.select = False
    for e in bm.edges:
        e.select = False
    # for e in border0:
    #     if e.is_valid:
    #         e.select = True

    if False:
        # bmesh.ops.mirror(bm, geom=bm.verts, axis="X", merge_dist=0.0001)
        bmesh.ops.symmetrize(bm, input=bm.faces, direction="Y", dist=0.001)
        bmesh.ops.remove_doubles(bm, verts=bm.verts, dist=0.0001)

        # remove wire edges
        bmesh.ops.delete(bm, geom=[e for e in bm.edges[:] if e.is_wire], context="EDGES")
